import os.path

import matplotlib.pyplot as plt
import numpy as np
import pickle
import arithmetics
from graph import act_t, Node, NodePtr

np.set_printoptions(precision=4, suppress=True, linewidth=np.inf)


class Goofspiel(object):
    def __init__(self):
        # non-iterables
        self.CARD_NUM: int = 3
        self.upcard: list[int] = [1, 1, 1, 1]
        self.weight: dict[str, list[float]] = {"A": [5.00, 1.33, 2.71, 1.80], "B": [4.12, 6.28, 3.33, 1.92]}
        self.root: NodePtr = NodePtr(0)
        self.bfs: list[list[NodePtr]] = []

        # instants
        self.A: dict[NodePtr, np.ndarray] = {}  # self.A[info_set][i, j]
        self.B: dict[NodePtr, np.ndarray] = {}  # self.B[info_set][i, j]

        # records
        self.reg: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # reg[player][info_set][act, t]
        self.sig: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # sig[player][info_set][act, t]
        self.cfv: dict[NodePtr, np.ndarray] = {}  # cfv[node][player, t]

    def tree_init(self):
        self.root = Node.new()
        self.root.o.parent = NodePtr(0)
        self.root.o.access = np.array([0, 0])
        self.root.o.actA = np.arange(self.CARD_NUM) + 1
        self.root.o.actB = np.arange(self.CARD_NUM) + 1

        self.bfs = []
        parent = [self.root]
        child = []
        while True:
            self.bfs.append(parent)
            for n in parent:
                n.o.child = np.resize(n.o.child, [len(n.o.actA), len(n.o.actB)])
                for i, a in enumerate(n.o.actA):
                    for j, b in enumerate(n.o.actB):
                        ch = Node.new()
                        n.o.child[i, j] = ch
                        ch.o.parent = n
                        ch.o.access = np.array([a, b])
                        ch.o.actA = n.o.actA[n.o.actA != a]
                        ch.o.actB = n.o.actB[n.o.actB != b]
                        child.append(ch)
            if len(child) == 0:
                break
            else:
                parent = child
                child = []

    def sigma_init(self, T: int = 200):
        for generation in self.bfs[0:-2]:
            for n in generation:
                self.reg["A"][n] = np.empty([len(n.o.actA), T], dtype=float)
                self.reg["B"][n] = np.empty([len(n.o.actB), T], dtype=float)
                self.sig["A"][n] = np.empty([len(n.o.actA), T], dtype=float)
                self.sig["B"][n] = np.empty([len(n.o.actB), T], dtype=float)
        for generation in self.bfs[0:-2]:
            for n in generation:
                self.reg["A"][n][:, 0] = arithmetics.rand_f(1, len(n.o.actA) - 1)[0] * 5
                self.reg["B"][n][:, 0] = arithmetics.rand_f(1, len(n.o.actB) - 1)[0] * 5
                self.sig["A"][n][:, 0] = self.reg["A"][n][:, 0] / self.reg["A"][n][:, 0].sum()
                self.sig["B"][n][:, 0] = self.reg["B"][n][:, 0] / self.reg["B"][n][:, 0].sum()

    def cfv_init(self, T: int = 200):
        for generation in self.bfs:
            for n in generation:
                self.cfv[n] = np.empty([2, T], dtype=float)
        for generation in self.bfs[0:-2]:
            for n in generation:
                self.A[n] = np.empty([len(n.o.actA), len(n.o.actB)], dtype=float)
                self.B[n] = np.empty([len(n.o.actA), len(n.o.actB)], dtype=float)
        # leaves:
        for n in self.bfs[-1]:
            h = n.o.h()
            u, v = 0, 0
            for k, acts in enumerate(h):
                a, b = acts
                if a > b:
                    u += self.upcard[k] * self.weight["A"][k]
                    v -= self.upcard[k] * self.weight["B"][k]
                elif a < b:
                    u -= self.upcard[k] * self.weight["A"][k]
                    v += self.upcard[k] * self.weight["B"][k]
            self.cfv[n][:, 0] = np.array([u, v], dtype=float)
        for n in self.bfs[-2]:
            self.cfv[n][:, 0] = self.cfv[n.o.get_child(0, 0)][:, 0].copy()

    def cfv_refresh(self, t: int):
        for generation in self.bfs[-1:-3:-1]:
            for n in generation:
                self.cfv[n][:, t] = self.cfv[n][:, 0]
        for generation in self.bfs[-3::-1]:
            for n in generation:
                for i, a in enumerate(n.o.actA):
                    for j, b in enumerate(n.o.actB):
                        self.A[n][i, j], self.B[n][i, j] = self.cfv[n.o.get_child(i, j)][:, t]
                self.cfv[n][0, t] = self.sig["A"][n][:, t].T @ self.A[n] @ self.sig["B"][n][:, t]
                self.cfv[n][1, t] = self.sig["A"][n][:, t].T @ self.B[n] @ self.sig["B"][n][:, t]
    # New idea: Ditch iterables, rely solely on records! Because this can tackle overwrites! 2022-1121-2330
    # Problem: leaves, trivial nodes? -> Anyway, don't care about compression!

    def sigma_update(self, t: int):
        for generation in self.bfs[0:-2]:
            for n in generation:
                x = self.sig["A"][n][:, t - 1]
                y = self.sig["B"][n][:, t - 1]
                u = self.A[n] @ y
                v = self.B[n].T @ x
                r = {"A": u - u.T @ x, "B": v - v.T @ y}
                r["A"][r["A"] < 0] *= 0.05
                r["B"][r["B"] < 0] *= 0.05
                self.reg["A"][n][:, t] = self.reg["A"][n][:, t - 1] + r["A"]
                self.reg["B"][n][:, t] = self.reg["B"][n][:, t - 1] + r["B"]
                self.reg["A"][n][self.reg["A"][n][:, t] < 0, t] = 0
                self.reg["B"][n][self.reg["B"][n][:, t] < 0, t] = 0
                self.sig["A"][n][:, t] = self.reg["A"][n][:, t] / self.reg["A"][n][:, t].sum()
                self.sig["B"][n][:, t] = self.reg["B"][n][:, t] / self.reg["B"][n][:, t].sum()


def split_to_stackers(A: np.ndarray):
    M, N = A.shape
    y = []
    for i in range(M):
        y.append(A[i, :])
    return y


# noinspection PyUnreachableCode
def init():
    np.random.seed(133484)
    game = Goofspiel()
    game.tree_init()

    T = 32001
    game.sigma_init(T)
    game.cfv_init(T)
    game.cfv_refresh(0)
    for t in range(1, T):
        if t % 1000 == 0:
            print(t)
        game.sigma_update(t)
        game.cfv_refresh(t)

    if False:
        print("cfv_leaf:")
        for n in game.bfs[-2]:
            print(n, ":", game.cfv[n][:, 0])
        print()
    if True:
        # plot regret and sigma
        for l in game.bfs[0:-2]:
            for n in l:
                fig: plt.Figure
                title = str(n.o.h())
                fig = plt.figure(title, [8, 6], 96)
                fig.add_axes([0.05, 0.55, 0.4, 0.4])
                fig.add_axes([0.55, 0.55, 0.4, 0.4])
                fig.add_axes([0.05, 0.05, 0.4, 0.4])
                fig.add_axes([0.55, 0.05, 0.4, 0.4])
                fig.axes[0].stackplot(np.arange(T), *split_to_stackers(game.reg["A"][n]), labels=n.o.actA)
                fig.axes[0].set_title(r'R["A"][node][act, t]')
                fig.axes[1].stackplot(np.arange(T), *split_to_stackers(game.reg["B"][n]), labels=n.o.actB)
                fig.axes[1].set_title(r'R["B"][node][act, t]')
                fig.axes[2].stackplot(np.arange(T), *split_to_stackers(game.sig["A"][n]), labels=n.o.actA)
                fig.axes[2].set_title(r'$\sigma$["A"][node][act, t]')
                fig.axes[3].stackplot(np.arange(T), *split_to_stackers(game.sig["B"][n]), labels=n.o.actB)
                fig.axes[3].set_title(r'$\sigma$["B"][node][act, t]')
                ax: plt.Axes
                for ax in fig.axes:
                    ax.grid()
                    ax.legend()
                fig.savefig(os.path.join("result", title + ".png"))
                plt.close(fig)

        # plot cfv

        # plot exploit-ability


if __name__ == "__main__":
    init()
